import { Resolver, Segment, Value, Variable } from '@stringifyrjs/lib';
import { type TResolversWithValue, type TTypeMap } from '@stringifyrjs/lib/src/TypeMap-Types';
import * as _ from 'lodash';
import { TypeMap } from './TypeMap';

class TypeMapGeneratorBase {
  constructor() {
    // Singleton for caching purposes
  }

  readonly typeMap = (typeMap: TResolversWithValue) => {
    const result: TTypeMap = {};

    for (const [resolver, value] of typeMap) {
      const variantTypeMapTree = TypeMapGenerator.resolverTypeMap(resolver, value);
      TypeMapGenerator.mergeTypeMapTree(result, variantTypeMapTree);
    }

    return result;
  };

  /**
   * todo comment
   * @param resolver
   * @param value
   */
  readonly resolverTypeMap = (resolver: string, value: string): TTypeMap => {
    const resolverSegments = Resolver.split(resolver);

    const stringValueVariables = Value.toVariables(value);

    const deep = resolverSegments.reverse().reduce((acc, segment) => {
      const result = {
        __def: {
          // The variables of a parent are always a superset of the
          // variables of a child
          // E.g if a child has a variable 'a' then a parent always has the variable 'a'
          __variables: _.cloneDeep(acc.__def.__variables ?? {}),
        },
      } as TTypeMap;

      function addResolverParent(parent: string) {
        _.merge(result.__def, {
          resolvers: {
            [parent]: {},
            ..._.mapKeys(acc.__def.resolvers, (val, childResolver) => {
              return Segment.join(parent, childResolver);
            }),
          },
        });
      }

      // Add the base segment
      result[segment] = _.cloneDeep(acc);
      addResolverParent(segment);

      // Add the variable name segment
      const variable = Segment.toVariable(segment);
      if (variable) {
        // Add the missing variable since this is the
        // un-valued variable segment
        _.merge(
          result.__def.__variables,
          TypeMap.recordToTypeMapVariableTypes({
            [variable.name]: variable.value,
          }),
        );

        const templateSegment = Variable.toTemplateSegment(variable);
        result[templateSegment] = {
          ..._.cloneDeep(acc),
          __def: _.merge(acc.__def ?? {}, {
            __variables: TypeMap.recordToTypeMapVariableTypes({
              [variable.name]: variable.value,
            }),
          }),
        };

        addResolverParent(templateSegment);
      }

      return result;
    }, {
      __def: {
        __variables: TypeMap.recordToTypeMapVariableTypes(stringValueVariables),
      },
    } as TTypeMap);

    return deep;
  };

  /**
   * todo comment
   */
  readonly mergeTypeMapTree = (tmt1: TTypeMap, tmt2: TTypeMap): TTypeMap => {
    return _.mergeWith(tmt1, tmt2, function(objValue, srcValue, key) {
      if (key === '__variables') {
        return _.mergeWith(objValue, srcValue, function(varObjValue, srcObjValue) {
          return _.uniq(srcObjValue.concat(varObjValue ?? [])).sort((a, b) => a < b ? -1 : 1);
        });
      }
    });
  };
}

export const TypeMapGenerator = new TypeMapGeneratorBase();
